package net.mawis.imip;

import java.io.*;
import java.util.*;

/**
 * This is an IMPacket that represents a message
 *
 * @author Matthias Wimmer &lt;m@tthias.net&gt;
 * @version 0.1.0
 */
public class IMMessage extends IMPacket {
    /**
     * Construct a new IMMessage
     *
     * @param encoding encoding of the packet
     */
    protected IMMessage(String encoding) {
	super("MESG", encoding);
    }

    /**
     * Read an IMMessage from an InputStream
     *
     * @param stream the InputStream to read from
     * @param encoding encoding of the packet
     */
    protected IMMessage(InputStream stream, String encoding) throws IOException {
	super(stream, "MESG", encoding);
    }

    /**
     * Get a new IMMessage
     *
     * @param userTo the user to which this message is directed
     * @param userFrom the user that sends this message
     * @param encoding encoding of the packet
     * @return new IMMessage packet
     */
    public static IMMessage getNewIMMessage(String userTo, String userFrom, String encoding) {
	    IMMessage newPacket = new IMMessage(encoding);
	    newPacket.addHeader("ACK-Type", "errors-only");
	    newPacket.addHeader("From", userFrom);
	    newPacket.addHeader("To", userTo);
	    return newPacket;
    }

    /**
     * Get the sender of a message
     *
     * @return the sender
     */
    public String getFrom() {
	String from = getHeader("From");
	if (from == null) {
	    return null;
	}

	int spaceIndex = from.indexOf(' ');
	if (spaceIndex>=0) {
	    return from.substring(0, spaceIndex);
	}
	return from;
    }

    /**
     * Get the date and time of the message
     *
     * @return the date and time of the message
     */
    public Date getDate() {
	String time = getHeader("Time");

	// if message has no time header we return the current time
	if (time == null) {
	    return new Date();
	}

	int year, month, day, hour, minute, second;
	try {
	    year = Integer.parseInt(time.substring(0, 4));
	    month = Integer.parseInt(time.substring(5, 7))-1;
	    day = Integer.parseInt(time.substring(8, 10));
	    hour = Integer.parseInt(time.substring(11, 13));
	    minute = Integer.parseInt(time.substring(14, 16));
	    second = Integer.parseInt(time.substring(17, 19));

	    Calendar calendar = Calendar.getInstance();
	    calendar.set(year, month, day, hour, minute, second);
	    return calendar.getTime();
	} catch (NumberFormatException nfe) {
	    return new Date();
	}
    }
}
