package net.mawis.imip;

import java.util.*;
import java.io.*;

/**
 * IMPacket represents a packet that has been received from the IMIP server
 * or that will be sent to the IMIP server
 *
 * @author Matthias Wimmer &lt;m@tthias.net&gt;
 * @version 0.1.0
 */
public class IMPacket {
    private String type;
    private Hashtable headers = new Hashtable();
    private String body = "";
    protected String encoding;

    /**
     * This constructor creates a new packet of the specified type to be
     * sent later to the server.
     *
     * @param type type of the packet to be created
     * @param encoding encoding of the packet
     */
    protected IMPacket(String type, String encoding) {
	this.type = type;
	this.encoding = encoding;
    }

    /**
     * Get a new IMPacket with the specified type
     *
     * @param type type of the packet to be created
     * @param encoding encoding of the packet
     * @return the new IMPacket
     */
    public static IMPacket getNewIMPacket(String type, String encoding) {
	if (type.equals("MESG")) {
	    return new IMMessage(encoding);
	} else if (getMainTypeFromString(type).equals("LIST")) {
	    return new IMSubscription(type, encoding);
	} else if (getMainTypeFromString(type).equals("STAT")) {
	    return new IMPresence(type, encoding);

	}
	return new IMPacket(type, encoding);
    }

    /**
     * Read an IMPacket from an InputStream
     *
     * @param stream stream to read the packet from
     * @param encoding encoding of headers and body
     * @return the read IMPacket
     */
    public static IMPacket getIMPacketFromStream(InputStream stream, String encoding) throws IOException {
	String type = "";
	int nextChar = stream.read();

	while (nextChar != '\n') {
	    if (nextChar == -1) {
		throw new IOException("Disconnected");
	    }
	    if (nextChar != '\r') {
		type = type+(char)nextChar;
	    }
	    nextChar = stream.read();
	}

	if (type.equals("MESG")) {
	    return new IMMessage(stream, encoding);
	} else if (getMainTypeFromString(type).equals("LIST")) {
	    return new IMSubscription(stream, type, encoding);
	} else if (getMainTypeFromString(type).equals("STAT")) {
	    return new IMPresence(stream, type, encoding);
	}
	return new IMPacket(stream, type, encoding);
    }

    /**
     * This constructor reads a packet from an InputStream
     *
     * @param stream stream to read the packet from the stream
     * @param type type of the packet
     * @param encoding encoding of headers and body
     */
    protected IMPacket(InputStream stream, String type, String encoding) throws IOException {
	this.encoding = encoding;
	this.type = type;

	// read second line (size of headers and body)
	int nextChar = stream.read();
	String pktLength = "";
	while (nextChar != '\n') {
	    if (nextChar == -1) {
		throw new IOException("Disconnected");
	    }
	    if (nextChar != '\r') {
		pktLength = pktLength+(char)nextChar;
	    }
	    nextChar = stream.read();
	}

	// read headers and body
	int length;
	try {
	    length = Integer.parseInt(pktLength);
	} catch (NumberFormatException nfe) {
	    throw new IOException("Protocol error. Packet length expected");
	}

	if (length == 0) return;

	byte[] buffer = new byte[length];
	int readBytes = 0;
	while (readBytes < length) {
	    int temp = stream.read(buffer, readBytes, length-readBytes);
	    if (temp == -1) {
		throw new IOException("Disconnected");
	    }
	    readBytes += temp;
	}

	// get headers and body out of the read byte array
	BufferedReader bfReader = new BufferedReader(new InputStreamReader((new ByteArrayInputStream(buffer, 0, length)),encoding));
	// headers
	String headerLine = bfReader.readLine();
	if (headerLine == null) {
	    headerLine = "";
	}
	while (!headerLine.equals("")) {
	    int colon = headerLine.indexOf(':');
	    String headerName = "";
	    String headerContent = "";
	    if (colon >= 0) {
		headerName = headerLine.substring(0, colon);
		headerContent = headerLine.substring(colon+1).trim();
	    } else {
		headerName = headerLine.trim();
	    }
	    headers.put(headerName, headerContent);
	    
	    headerLine = bfReader.readLine();
	    if (headerLine == null) {
		headerLine = "";
	    }
	}
	// body
	try {
	    while (bfReader.ready()) {
		body = body+(char)bfReader.read();
	    }
	} catch (IOException ioe) {}
    }

    /**
     * This method gets the binary representation of the packet
     *
     * @return binary representation of the packet
     */
    public byte[] getBytes() {
	try {
	    return toString().getBytes(encoding);
	} catch (UnsupportedEncodingException uee) {
	    return toString().getBytes();
	}
    }

    /**
     * Return the packet as a String
     *
     * @return string represenation of the packet
     */
    public String toString() {
	String content = "";
	Enumeration elements = headers.keys();
	while (elements.hasMoreElements()) {
	    String key = (String)elements.nextElement();
	    content = content+key+": "+(String)headers.get(key)+"\r\n";
	}
	content = content+"\r\n"+body;
	int contentSize = 0;
	try {
	    contentSize = content.getBytes(encoding).length;
	} catch (UnsupportedEncodingException uee) {
	    contentSize = content.getBytes().length;
	}

	return type+"\r\n"+(new Integer(contentSize)).toString()+"\r\n"+content;
    }

    /**
     * Add a header line
     *
     * @param name Name of the header
     * @param value Value of the header
     */
    public void addHeader(String name, String value) {
	headers.put(name, value);
    }

    /**
     * Set the body of a packet
     *
     * @param body Body of the packet
     */
    public void setBody(String body) {
	this.body = body;
    }

    /**
     * Get the body of a packet
     *
     * @return body of the packet
     */
    public String getBody() {
	return body;
    }

    /**
     * Get the type of a packet
     *
     * @return packet type
     */
    public String getType() {
	return type;
    }

    /**
     * Get a header of a packet
     *
     * @param name name of the header
     * @return value of the header
     */
    public String getHeader(String name) {
	return (String)headers.get(name);
    }

    /**
     * Get the main type out of a type
     *
     * @param type the complete type
     * @return the main type
     */
    protected static String getMainTypeFromString(String type) {
	if (type == null) {
	    return "";
	}
	
	int sep = type.indexOf(' ');
	if (sep > 0) {
	    return type.substring(0, sep);
	}
	return type;
    }

    /**
     * Get the main type of this packet
     *
     * @return the main type of this packet
     */
    public String getMainType() {
	return getMainTypeFromString(type);
    }

    /**
     * Get the sub type out of a type
     *
     * @param type the complete type
     * @return the sub type
     */
    protected static String getSubTypeFromString(String type) {
	if (type == null) {
	    return "";
	}
	
	int sep = type.indexOf(' ');
	if (sep > 0) {
	    return type.substring(sep+1).trim();
	}
	return "";
    }
    
    /**
     * Get the sub type of this packet
     *
     * @return the sub type of this packet
     */
    public String getSubType() {
	return getSubTypeFromString(type);
    }
}
