package net.mawis.imip;

import java.util.*;

/**
 * This object represents the contact list of an user. It keeps track of
 * subscriptions and presences of all the users.
 *
 * @author Matthias Wimmer &lt;m@tthias.net&gt;
 * @version 0.1.0
 */
public class Roster implements PresenceListener {
    private Hashtable items;
    private Vector rosterListeners;

    /**
     * Constructor for the Roster class
     */
    public Roster() {
	items = new Hashtable();
    }
    
    /**
     * A new IMPresence has been received by the ServerConnection and
     * Roster is notified about that with this method.
     * Should not be called by the user.
     *
     * @param newPresence the IMPresence packet that has been received
     */
    public void newPresence(IMPresence newPresence) {
	RosterItem item = getItemByID(newPresence.getSenderID(), true);
	item.setPresenceState(newPresence.getPresenceState());
	notifyRosterListeners(item, false);
    }

    /**
     * Set the initial items that have been received within the LOGN
     * packet
     *
     * @param initialItems the initialItems from the LOGN packet
     */
    void setInitialItems(String initialItems) {
	String nextItem;
	String itemNick;
	if (initialItems != null) {
	    initialItems = initialItems.trim();
	}
	
	while (initialItems != null && !initialItems.equals("")) {
	    int pos = initialItems.indexOf(',');
	    if (pos>=0) {
		nextItem = initialItems.substring(0, pos).trim();
		initialItems = initialItems.substring(pos+1).trim();
	    } else {
		nextItem = initialItems.trim();
		initialItems = null;
	    }
	    
	    pos = nextItem.indexOf(' ');
	    if (pos>=0) {
		itemNick = nextItem.substring(pos+1).trim();
		nextItem = nextItem.substring(0, pos).trim();
	    } else {
		itemNick = nextItem;
	    }
	    if (itemNick.charAt(0)=='"' && itemNick.charAt(itemNick.length()-1)=='"') {
		itemNick = itemNick.substring(0, itemNick.length()-1).substring(1);
	    }

	    if (!nextItem.equals("")) {
		RosterItem itemToChange = getItemByID(nextItem, true);
		boolean moreChanges = (nextItem == null || nextItem.equals(""))?false:true;
		itemToChange.setNick(itemNick);
		notifyRosterListeners(itemToChange, moreChanges);
	    }
	}
    }

    /**
     * Get a roster item by the ID
     *
     * @param id the user's ID
     * @param create create item if it doesn't exist
     * @return the RosterItem (if found or created)
     */
    private RosterItem getItemByID(String id, boolean create) {
	RosterItem item = (RosterItem)items.get(id);

	if (create && item == null) {
	    item = new RosterItem(id);
	    items.put(id, item);
	}

	return (RosterItem)items.get(id);
    }

    /**
     * Get a roster item by the ID
     *
     * @param id the user's ID
     * @return the RosterItem (if any)
     */
    public RosterItem getItemByID(String id) {
	return getItemByID(id, false);
    }

    /**
     * get String representation of this roster
     *
     * @return String representation
     */
    public String toString() {
	String result = "";

	Enumeration elements = items.keys();
	while (elements.hasMoreElements()) {
	    String key = (String)elements.nextElement();
	    result += getItemByID(key).toString() + "\n";
	}

	return result;
    }

    /**
     * add a RosterChangeListener
     *
     * @param listener the new RosterChangeListener
     */
    public void addRosterChangeListener(RosterChangeListener listener) {
	if (rosterListeners == null) {
	    rosterListeners = new Vector();
	}
	rosterListeners.addElement(listener);
    }

    /**
     * notify all registered instances of RosterListener about a change
     *
     * @param changedItem the RosterItem that has changed its state
     * @param more true if more items will change soon
     */
    public void notifyRosterListeners(RosterItem changedItem, boolean more) {
	if (rosterListeners == null) {
	    return;
	}
	Enumeration listeners = rosterListeners.elements();
	while (listeners.hasMoreElements()) {
	    Object listener = listeners.nextElement();
	    if (listener instanceof RosterChangeListener) {
		((RosterChangeListener)listener).rosterItemChanged(changedItem, more);
	    }
	}
    }
}
