package net.mawis.imip;

import java.net.*;
import java.io.*;
import java.security.*;
import java.util.*;

/**
 * ServerConnection represents the connection to an IMIP server
 *
 * @author Matthias Wimmer &lt;m@tthias.net&gt;
 * @version 0.1.0
 */
public class ServerConnection extends Thread {
    private String host;
    private int port;
    private String user;
    private String passwd;
    private String encoding = "UTF-8";
    
    private Socket socket;
    private InputStream inputStream;
    private OutputStream outputStream;

    private ConnectionState connectionState;
    private int idCounter = 1001;
    private int keepAliveInterval = 0;
    private int keepAliveTimer = 0;
    private boolean keepRunning = true;
    private String hexDigits = "0123456789abcdef";
    private Vector stateListeners;
    private Vector messageListeners;
    private Vector subscriptionListeners;
    private Vector presenceListeners;
    private String disconnectRequestID;
    private boolean disconnectRequested = false;
    private String lognID = null;
    private Roster roster;
    

    /**
     * This constructor creates a new instance of ServerConnections and sets
     * the information where to connect to as which user.
     *
     * @param host host to connect to
     * @param port port to connect to
     * @param user username
     * @param passwd password
     */
    public ServerConnection(String host, int port, String user, String passwd) {
	this.host = host;
	this.port = port;
	this.user = user;
	this.passwd = passwd;

	connectionState = new ConnectionState();
	roster = new Roster();
	addPresenceListener(roster);
    }

    /**
     * Connect to the IMIP server and log in
     */
    public void connect() throws IOException {
	// connect to the server
	socket = new Socket(host, port);

	// get input and output stream
	inputStream = socket.getInputStream();
	outputStream = socket.getOutputStream();

	// update connection state
	changeState(ConnectionState.STATE_AUTHENTICATING, false);

	// start listening on the input stream
	keepRunning = true;
	start();

	// send the HELO packet
	IMPacket imPacket = IMPacket.getNewIMPacket("HELO", encoding);
	imPacket.addHeader("Protocol", "IMIP/1.0");
	sendPacketWithID(imPacket);
    }

    /**
     * Send an IMPacket
     *
     * @param packet the packet to be sent
     */
    public void sendPacket(IMPacket packet) {
	byte[] binPacket = packet.getBytes();

	if (outputStream == null) {
	    return;
	}

	try {
	    outputStream.write(binPacket, 0, binPacket.length);
	} catch (IOException ioe) {
	    changeState(ConnectionState.STATE_ERROR, true);
	}
    }

    /**
     * Send an IMPacket after adding an ID value
     *
     * @param packet the packet to be sent
     * @return the ID that has been assigned to this packet
     */
    public String sendPacketWithID(IMPacket packet) {
	String idString = (new Integer(idCounter++)).toString();

	packet.addHeader("ID", idString);
	sendPacket(packet);
	return idString;
    }

    /**
     * Poll the input stream for new packets
     */
    public void run() {
	while (keepRunning) {
	    // is a new packet available?
	    try {
		if (inputStream.available() > 0) {
		    IMPacket readPacket = IMPacket.getIMPacketFromStream(inputStream, encoding);
		    processPacket(readPacket);
		}
	    } catch (IOException ioe) {
		changeState(ConnectionState.STATE_ERROR, true);
	    }

	    // sleep a bit
	    try {
		Thread.sleep(2000);
	    } catch (Exception e) {}

	    // send new PING packet?
	    if (keepAliveInterval > 0) {
		keepAliveTimer -= 2;
		if (keepAliveTimer <= 0) {
		    IMPacket kaPacket = IMPacket.getNewIMPacket("PING", encoding);
		    sendPacketWithID(kaPacket);
		    keepAliveTimer = keepAliveInterval;
		}
	    }
	}
    }

    /**
     * process a received packet
     *
     * @param packet the packet that has been received
     */
    private void processPacket(IMPacket packet) {
	String typeMain = packet.getMainType().toUpperCase();
	String typeSub = packet.getSubType().toUpperCase();

	if (typeMain.equals("HELO")) {
	    // we're told how we can authenticate
	    authenticate(packet);

	    // we have to set up the keep alives
	    try {
		keepAliveInterval = Integer.parseInt(packet.getHeader("Keep-Alive"));
	    } catch (NumberFormatException nfs) {
	    } catch (NullPointerException npe) {}
	    keepAliveTimer = keepAliveInterval;
	} else if (typeMain.equals("LOGN")) {
	    changeState(ConnectionState.STATE_CONNECTED, false);
	    roster.setInitialItems(packet.getHeader("Buddy"));
	} else if (typeMain.equals("ACK")) {
	    processACK(packet);
	} else if (typeMain.equals("STAT")) {
	    if (packet instanceof IMPresence) {
		newPresence((IMPresence)packet);
	    } else {
		System.out.println("Bug: Received STAT packet that is no instance of IMPresence");
		System.out.println(packet.toString());
	    }
	} else if (typeMain.equals("MESG")) {
	    if (packet instanceof IMMessage) {
		newMessage((IMMessage)packet);
	    } else {
		System.out.println("Bug: Received MESG packet that is no instance of IMMessage");
		System.out.println(packet.toString());
	    }
	} else if (typeMain.equals("LIST")) {
	    if (packet instanceof IMSubscription) {
		newSubscription((IMSubscription)packet);
	    } else {
		System.out.println("Bug: Received LIST packet that is no instance of IMSubscription");
		System.out.println(packet.toString());
	    }
	} else {
	    System.out.println("Unknown packet from server:");
	    System.out.println(packet.toString());
	}
    }

    /**
     * process a received IMPacket of main type ACK
     *
     * @param packet the received packet
     */
    private void processACK(IMPacket packet) {
	String reference = packet.getHeader("Reference");

	if (!packet.getType().equals("ACK 600")) {
	    System.out.println(packet.toString());
	}

	if (lognID != null && reference != null) {
	    if (reference.equals(lognID) && packet.getType().charAt(4)!='6') {
		changeState(ConnectionState.STATE_ERROR, true);
	    }
	}

	if (disconnectRequested) {
	    if (reference != null && disconnectRequestID != null) {
		if (reference.equals(disconnectRequestID)) {
		    changeState(ConnectionState.STATE_DISCONNECTED, true);
		}
	    }
	}
    }

    /**
     * send authentication packet
     *
     * @param heloPacket the packet with information about how we can login
     */
    private void authenticate(IMPacket heloPacket) {
	String authType = heloPacket.getHeader("Auth-Type");
	if (authType == null) {
	    System.out.println("No Auth-Type header ... don't know how to authenticate");
	    changeState(ConnectionState.STATE_ERROR, true);
	    return;
	}
	if (!authType.equals("imip-md5")) {
	    System.out.println("Unsupported Auth-Type "+authType);
	    changeState(ConnectionState.STATE_ERROR, true);
	    return;
	}
	IMPacket lognPacket = IMPacket.getNewIMPacket("LOGN "+user, encoding);
	lognPacket.addHeader("Client", "JavaIMIP");
	lognPacket.addHeader("Auth-Type", authType);

	// process salt and password
	String secret = heloPacket.getBody().trim();
	secret = secret+passwd;
	MessageDigest md;
	try {
	    md = MessageDigest.getInstance("MD5");
	} catch (NoSuchAlgorithmException nsae) {
	    System.out.println("Coundn't get Digest: "+nsae.toString());
	    changeState(ConnectionState.STATE_ERROR, true);
	    return;
	}
	try {
	    md.update(secret.getBytes(encoding));
	} catch (Exception e) {
	    md.update(secret.getBytes());
	}
	byte[] binSecret = md.digest();
	secret = "";
	for (int i = 0; i<16; i++) {
	    int hexDigit = binSecret[i];
	    if (hexDigit<0) {
		hexDigit = 256+hexDigit;
	    }
	    secret = secret+hexDigits.charAt(hexDigit/16);
	    secret = secret+hexDigits.charAt(hexDigit%16);
	}
	lognPacket.setBody(secret);

	lognID = sendPacketWithID(lognPacket);
    }

    /**
     * disconnect from the server
     */
    public void disconnect() {
	keepAliveInterval = 0;
	IMPacket imPacket = IMPacket.getNewIMPacket("DISC", encoding);
	disconnectRequestID = sendPacketWithID(imPacket);
	disconnectRequested = true;
    }

    /**
     * change the state of this connection, notify the listeners and
     * maybe disconnect the socket
     *
     * @param newState new state
     * @param boolean disconnect the socket
     */
    private void changeState(int newState, boolean disconnect) {
	connectionState.setState(newState);
	if (disconnect) {
	    keepRunning = false;
	    try {
		socket.close();
	    } catch (Exception ioe) { }
	}

	if (stateListeners != null) {
	    Enumeration listeners = stateListeners.elements();
	    while (listeners.hasMoreElements()) {
		Object listener = listeners.nextElement();
		if (listener instanceof StateChangeListener) {
		    ((StateChangeListener)listener).newConnectionState(connectionState);
		}
	    }
	}
    }

    /**
     * notify the IMMessageListeners about a new message that has been received
     *
     * @param newMessage the new message that has been received
     */
    private void newMessage(IMMessage newMessage) {
	if (messageListeners != null) {
	    Enumeration listeners = messageListeners.elements();
	    while (listeners.hasMoreElements()) {
		Object listener = listeners.nextElement();
		if (listener instanceof IMMessageListener) {
		    ((IMMessageListener)listener).newMessageReceived(newMessage);
		}
	    }
	}
    }

    /**
     * notify the SubscriptionListeners about a new subscription request
     *
     * @param packet the IMSubscription that has been received
     */
    private void newSubscription(IMSubscription packet) {
	if (subscriptionListeners != null) {
	    Enumeration listeners = subscriptionListeners.elements();
	    while (listeners.hasMoreElements()) {
		Object listener = listeners.nextElement();
		if (listener instanceof SubscriptionListener) {
		    ((SubscriptionListener)listener).newSubscriptionRequested(packet);
		}
	    }
	}
    }

    /**
     * notify the PresenceListeners about a new IMPresence packet
     *
     * @param packet the IMPresence packet that has been received
     */
    private void newPresence(IMPresence packet) {
	if (presenceListeners == null) {
	    return;
	}
	Enumeration listeners = presenceListeners.elements();
	while (listeners.hasMoreElements()) {
	    Object listener = listeners.nextElement();
	    if (listener instanceof PresenceListener) {
		((PresenceListener)listener).newPresence(packet);
	    }
	}
    }

    /**
     * add a StateChangeListener
     *
     * @param listener the new StateChangeListener
     */
    public void addStateChangeListener(StateChangeListener listener) {
	if (stateListeners == null) {
	    stateListeners = new Vector();
	}
	stateListeners.addElement(listener);
    }

    /**
     * add a IMMessageListener
     *
     * @param listener the new IMMessageListener
     */
    public void addIMMessageListener(IMMessageListener listener) {
	if (messageListeners == null) {
	    messageListeners = new Vector();
	}
	messageListeners.addElement(listener);
    }

    /**
     * add a SubscriptionListener
     *
     * @param listener the new SubscriptionListener
     */
    public void addSubscriptionListener(SubscriptionListener listener) {
	if (subscriptionListeners == null) {
	    subscriptionListeners = new Vector();
	}
	subscriptionListeners.addElement(listener);
    }

    /**
     * add a PresenceListener
     *
     * @param listener the new PresenceListener
     */
    public void addPresenceListener(PresenceListener listener) {
	if (presenceListeners == null) {
	    presenceListeners = new Vector();
	}
	presenceListeners.addElement(listener);
    }

    /**
     * get the associated Roster instance
     *
     * @return the associated Roster instance
     */
    public Roster getRoster() {
	return roster;
    }
}
